#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import OrderedDict

def parser_blank(infile) -> dict:
    """
    Parser that does nothing, for testing.
    """
    return OrderedDict()

def ignored(infile) -> dict:
    """
    Ignored function, for testing.
    """
    return OrderedDict()

def textparser_textblank(infile) -> dict:
    """
    Parser that does nothing, for testing.
    """
    return OrderedDict()

def rawparser_rawblank(infile) -> dict:
    """
    Parser that does nothing, for testing.
    """
    return OrderedDict()

def pathparser_pathblank(infile) -> dict:
    """
    Parser that does nothing, for testing.
    """
    return OrderedDict()

# expected result of importing
expected_import_parser = {
    'blank' : ('parser_blank', 'text', parser_blank),
    'textblank' : ('textparser_textblank', 'text', textparser_textblank),
    'rawblank' : ('rawparser_rawblank', 'raw', rawparser_rawblank),
    'pathblank' : ('pathparser_pathblank', 'path', pathparser_pathblank)
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv
import itertools
import yaml
import re
from typing import Dict, IO, List, Optional, Tuple, Union

def _istrue(tokens, name):
    """
    Helper that returns True if item named `name` is present in tokens
    and value is a YAML True value.
    """
    return name in tokens and yaml.safe_load(tokens[name]) == True

def svtextparser_csv_wide(
    logfile: IO[str], tokens: Dict[str, str], delimiter=","
) -> Tuple[List[str], List[Union[str, None]]]:
    """
    Parses a 'wide' comma-separated value (CSV) file in the following format
    (whitespace is stripped):
    ```
    heading1, heading2,...
    value1,   value2, ...
    ```
    to `([heading1, heading2, ...], [value1, value2, ...])`.

    Any missing values are replaced by `None` (by default). 
    Any additional value without a heading is given heading 
    `heading{column_index}` where column index starts from 1. If the token 
    `"parse_no_headings"` is set, all rows are used as
    values and all headings are generated.
    `parse_heading_strip` and `parse_value_strip` are the character sets used
    for stripping from both sides of headings values; if unset defaults to
    stripping whitespace. Set to `""` to strip no characters.
    `parse_value_missing` is  used as the value when value is missing; 
    defaults to `None`. Can be an empty string to ensure prefix/suffix are 
    always added.
    `heading_suffix` and `heading_prefix` are the suffix and prefix added to
    all headings; this is final step before storing so all other processing is
    performed before these are added to the heading.
    `value_suffix` and `value_prefix` are the suffix and prefix added to
    all values; this is final step before storing so all other processing is
    performed before these are added to the value but requires that the 
    value is not `None`.

    If multiple rows of values are present, a suffix `_row{row_index}` is added 
    to all subsequent headings, where `row_index` starts from 2 for the second 
    values row.

    The token `"parse_delimiter"` will override the `delimiter` parameter, which
    is the delimiter that separates records in the file.

    Blank lines are skipped.

    This is a `svtext`-type parser as it uses the `csv` module.
    """
    if "parse_delimiter" in tokens:
        delimiter = tokens["parse_delimiter"]
    reader = csv.reader(logfile, delimiter=delimiter)

    parse_heading_strip = tokens.get("parse_heading_strip") 
    parse_value_strip = tokens.get("parse_value_strip") 
    parse_value_missing = tokens.get("parse_value_missing") 

    heading_suffix = tokens.get("heading_suffix", "")
    heading_prefix = tokens.get("heading_prefix", "")
    value_suffix = tokens.get("value_suffix", "")
    value_prefix = tokens.get("value_prefix", "")

    if "parse_no_headings" in tokens:
        file_headings = []
    else:
        try:
            file_headings = [
                heading.strip(parse_heading_strip)
                for heading in next(reader)]
        except StopIteration:
            file_headings = []
    out_headings = []
    out_values = []

    rows = []
    for row in reader:
        if len(row) == 0:
            continue  # skip blank lines
        rows.append(row)

    if len(rows) == 0:
        rows.append([])  # this ensures we have at least 1 row

    for row_index, row in enumerate(rows, start=1):
        row_suffix = f"_row{row_index}"
        for column, (file_heading, value) in enumerate(
            itertools.zip_longest(file_headings, row), start=1
        ):
            if value is None:
                value = parse_value_missing
            else:
                value = value.strip(parse_value_strip)

            if file_heading is None:
                file_heading = f"heading{column}"
            if row_index == 1:
                out_heading = file_heading
            else:
                out_heading = file_heading + row_suffix
            out_headings.append(heading_prefix + out_heading + heading_suffix)
            out_values.append(value if value is None else \
                value_prefix + value + value_suffix)
    return (out_headings, out_values)


def svtextparser_tsv_wide(
    logfile: IO[str], tokens: Dict[str, str]
) -> Tuple[List[str], List[Union[str, None]]]:
    """
    Parses a 'wide' tab-separated value (TSV) file in the following format
    (extra whitespace is stripped):
    ```
    heading1\t heading2\t...
    value1\t   value2\t ...
    ```
    to `{heading1: value1, heading2: value2, ...}`.

    See `pathparser_csv_wide()`.
    """
    return svtextparser_csv_wide(logfile, tokens, delimiter="\t")


def svtextparser_csv_tall(
    logfile: IO[str], tokens: Dict[str, str], delimiter=","
) -> Tuple[List[str], List[Union[str, None]]]:
    """
    Parses a 'tall' comma-separated value (CSV) file in the following format
    (whitespace is stripped):
    ```
    heading1, value1
    heading2, value2
    ...
    ```
    to `([heading1, heading2, ...], [value1, value2, ...])`.

    A row with only a single item is assumed to contain a heading alone, the 
    value is missing and replaced by `None` (by default). If the token 
    `"parse_indiv_as_values"` is set (present and evaluates to a YAML `True`
    value, this behaviour is changed so that the item is the value and the 
    missing heading is given as `heading{row_index}`. If the token 
    `"parse_no_headings"` is set, all rows are used as values and all headings are 
    genarated.
    `parse_heading_strip` and `parse_value_strip` are the character sets used
    for stripping from both sides of headings values; if unset defaults to
    stripping whitespace. Set to `""` to strip no characters.
    `parse_value_missing` is  used as the value when value is missing; 
    defaults to `None`. Can be an empty string to ensure prefix/suffix are 
    always added.
    `heading_suffix` and `heading_prefix` are the suffix and prefix added to
    all headings; this is final step before storing so all other processing is
    performed before these are added to the heading.
    `value_suffix` and `value_prefix` are the suffix and prefix added to
    all values; this is final step before storing so all other processing is
    performed before these are added to the value but requires that the 
    value is not `None`.

    If multiple columns of values are present, a suffix `_col{col_index}` is added 
    to all subsequent headings, where `col_index` starts from 2 for the second 
    values column.

    The token `"parse_delimiter"` will override the delimiter parameter, which
    is the delimiter that separates records in the file.

    Blank lines are skipped.

    This is a `svtext`-type parser as it uses the `csv` module.
    """
    delimiter = tokens.get("parse_delimiter", delimiter)
    indiv_as_values = _istrue(tokens, "parse_indiv_as_values")
    no_headings = _istrue(tokens, "parse_no_headings")

    parse_heading_strip = tokens.get("parse_heading_strip") 
    parse_value_strip = tokens.get("parse_value_strip") 
    parse_value_missing = tokens.get("parse_value_missing") 

    heading_suffix = tokens.get("heading_suffix", "")
    heading_prefix = tokens.get("heading_prefix", "")
    value_suffix = tokens.get("value_suffix", "")
    value_prefix = tokens.get("value_prefix", "")

    reader = csv.reader(logfile, delimiter=delimiter)
    out_headings = []
    out_values = []

    for row_index, row in enumerate(reader, start=1):
        if len(row) == 0:
            row_index -= 1
            continue  # skip blank lines
        if no_headings or (len(row) == 1 and indiv_as_values):
            file_heading = f"heading{row_index}"
            values = row  # use all as values
        else:
            file_heading = row[0].strip(parse_heading_strip)
            values = row[1:]
            if len(values) == 0:
                values.append(None)  # this ensures we have at least 1 value

        for column, value in enumerate(values, start=1):
            if value is None:
                value = parse_value_missing
            else:
                value = value.strip(parse_value_strip)
            if column == 1:
                out_heading = file_heading
            else:
                out_heading = file_heading + f"_row{column}"
            out_headings.append(heading_prefix + out_heading + heading_suffix)
            out_values.append(value if value is None else \
                value_prefix + value + value_suffix)
    return (out_headings, out_values)


def svtextparser_tsv_tall(
    logfile: IO[str], tokens: Dict[str, str]
) -> Tuple[List[str], List[Union[str, None]]]:
    """
    Parses a 'tall' tab-separated value (TSV) file in the following format
    (extra whitespace is stripped):
    ```
    heading1\t value1
    heading2\t value2
    ...
    ```
    to `([heading1, heading2, ...], [value1, value2, ...])`.

    See `pathparser_csv_tall()`.
    """
    return svtextparser_csv_tall(logfile, tokens, delimiter="\t")

def textparser_regex(
    logfile: IO[str], tokens: Dict[str, str]
) -> Tuple[List[str], List[Union[str, None]]]:
    """
    Parses a generic text file using regular expressions to extract
    `heading: value` pairs from the file.
    
    Regular expressions are specified as tokens named `parse_regex{name}` 
    with the regular expression given as their values. Regular expressions
    can either have named groups `heading*` and `value*`, which are used
    to derive the `heading` and `value` from parsing. If multiple groups 
    starting with `heading` or `value` are present, the group names are sorted
    and the resulting `heading` or `value` is derived by joining all groups
    with `heading_sep` or `value_sep` (default: `""`). If an underscore (`_`)
    is present in the group name, then the contents after the underscore
    are used as the group value, instead of the captured string itself.
    
    If `heading*` is not present, 
    `{name}` in the regular expression is used as the `heading`, with up to
    two leading underscores stripped from this (so `parse_regex_{name}` and
    `parse_regex__{name}` will both give `{name}` as the heading). If no named 
    group `value*` is not present then the last capturing group in the regular 
    expression is used as the `value` (at least one capturing group must be
    specified in the regular expression). 

    `heading_suffix` and `heading_prefix` are the suffix and prefix added to
    all headings; this is final step before storing so all other processing is
    performed before these are added to the heading.
    `value_suffix` and `value_prefix` are the suffix and prefix added to
    all values; this is final step before storing so all other processing is
    performed before these are added to the value.

    In general, the regular expressions are processed for each line
    in the file looking for matches using the `re.match` function, but there 
    are additional options to change this behaviour.

    The following additional parameters can be provided to the parser. They
    should be specified as YAML-style boolean options to activate:

    `parse_multiline`: Apply regular expression as multiline regular expression.
    Activates the `MULTILINE` flag and uses the entire contents of file as
    a single string when matching. Note: This will trigger loading of the
    entire file into memory.

    `parse_ignorecase`: Activate `IGNORECASE` flag when matching regular 
    expressions.

    `parse_unicode`: Activate `UNICODE` flag when matching regular 
    expressions.

    `parse_search`: Use the `re.search` function to find matches.

    `parse_findall`: Find all matches for a regular expression, rather than just
    the first (uses `re.finditer` to find matches).
    """

    re_flags = 0
    use_multiline = _istrue(tokens, "parse_multiline")
    ignore_case = _istrue(tokens, "parse_ignorecase")
    use_unicode = _istrue(tokens, "parse_unicode")
    parse_search = _istrue(tokens, "parse_search")
    parse_findall = _istrue(tokens, "parse_findall")

    heading_sep = tokens.get("heading_sep", "")
    heading_suffix = tokens.get("heading_suffix", "")
    heading_prefix = tokens.get("heading_prefix", "")
    value_sep = tokens.get("value_sep", "")
    value_suffix = tokens.get("value_suffix", "")
    value_prefix = tokens.get("value_prefix", "")

    if use_multiline: re_flags |= re.MULTILINE
    if ignore_case: re_flags |= re.IGNORECASE
    if use_unicode: re_flags |= re.UNICODE

    regex_param_re = re.compile("^parse_regex_{0,2}(.*)$")

    re_list = []
    for token, value in tokens.items():
        token_match = regex_param_re.match(token)
        if token_match:
            re_heading = token_match.group(1)
            re_compiled = re.compile(value, re_flags)
            re_list.append((re_heading, re_compiled))
    
    if use_multiline:
        input_str = ""
        for line in logfile:
            input_str+=line
        lines = (input_str,)
    else:
        lines = logfile
    
    out_headings = []
    out_values = []
    for line in lines:
        for re_heading, re_compiled in re_list:
            if parse_findall:
                re_matches = re_compiled.finditer(line)
            else:
                if parse_search:
                    re_match = re_compiled.search(line)
                else:
                    re_match = re_compiled.match(line)
                re_matches = (re_match,)
            for re_match in re_matches:
                if re_match is None: continue
                match_groups = re_match.groupdict()
                re_tokens_headings = {}
                re_tokens_values = {}
                for key,value in match_groups.items():
                    if "_" in key:
                        token = key.split("_",1)[1]
                    else:
                        token = value
                    if key.startswith("heading"):
                        re_tokens_headings[key] = token
                    if key.startswith("value"):
                        re_tokens_values[key] = token

                if re_tokens_headings:
                    out_heading = heading_sep.join(re_tokens_headings[key] for key in sorted(re_tokens_headings.keys()))
                else:
                    out_heading = re_heading
                if re_tokens_values:
                    out_value = value_sep.join(re_tokens_values[key] for key in sorted(re_tokens_values.keys()))
                else:
                    if len(re_match.groups())<1:
                        raise Exception(f"{re_heading}: regular expression must have at least one capturing group")
                    out_value = re_match.groups()[-1]
                
                out_headings.append(heading_prefix + out_heading + heading_suffix)
                out_values.append(value_prefix + out_value + value_suffix)
    return (out_headings, out_values)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
from collections import OrderedDict

from abrije.abrije import Abrije


class TestAbrijeClass:
    def test_add_parser(self):
        abrije_ = Abrije()  # test object
        parser_null = lambda handle: None  # dummy parsers for testing
        rawparser_null = lambda handle: None
        pathparser_null = lambda path: None

        # test add parser of default type: text
        abrije_.add_parser("parser_test", parser_null)
        expected_add_parser = {"test": ("parser_test", "text", parser_null)}
        assert abrije_.parsers == expected_add_parser

        # test adding parser of various types: text, raw, path
        abrije_.add_parser("textparser_test1", parser_null)
        abrije_.add_parser("svtextparser_test1a", parser_null)
        abrije_.add_parser("rawparser_test2", rawparser_null)
        abrije_.add_parser("pathparser_test3", pathparser_null)
        expected_add_parser.update(
            {
                "test1": ("textparser_test1", "text", parser_null),
                "test1a": ("svtextparser_test1a", "svtext", parser_null),
                "test2": ("rawparser_test2", "raw", rawparser_null),
                "test3": ("pathparser_test3", "path", pathparser_null),
            }
        )
        assert abrije_.parsers == expected_add_parser

        # test exception on adding unsupported type
        with pytest.raises(ValueError, match="unknown type"):
            abrije_.add_parser("strangeparser_test", parser_null)

        # test silent ignore parser of wrong name format
        prev_parsers = abrije_.parsers.copy()
        abrije_.add_parser("ignored", parser_null)
        abrije_.parsers
        assert abrije_.parsers == prev_parsers

        # test warning on add parser of same name
        # and no new parser added
        prev_parsers = abrije_.parsers.copy()
        with pytest.warns(UserWarning, match="same name"):
            abrije_.add_parser("rawparser_test", parser_null)
        assert abrije_.parsers == prev_parsers

        # test exception on adding non-callable func
        with pytest.raises(ValueError, match="not callable"):
            abrije_.add_parser("parser_test4", "a string")

    def test_import_parsers(self):

        # test import from module
        abrije_ = Abrije()  # test object
        import abrije.tests.parser_fortest  # module must be imported first

        abrije_.import_parsers(abrije.tests.parser_fortest)
        assert abrije_.parsers == abrije.tests.parser_fortest.expected_import_parser

        abrije_ = Abrije()  # reset test object
        import abrije.tests.parser_fortest_warn

        with pytest.warns(UserWarning, match="same name"):
            abrije_.import_parsers(abrije.tests.parser_fortest_warn)
        assert (
            abrije_.parsers == abrije.tests.parser_fortest_warn.expected_import_parser
        )

        abrije_ = Abrije()  # reset test object
        import abrije.tests.parser_fortest_err1

        with pytest.raises(ValueError, match="unknown type"):
            abrije_.import_parsers(abrije.tests.parser_fortest_err1)
        assert (
            abrije_.parsers == abrije.tests.parser_fortest_err1.expected_import_parser
        )

        abrije_ = Abrije()  # reset test object
        import abrije.tests.parser_fortest_err2

        with pytest.raises(ValueError, match="not callable"):
            abrije_.import_parsers(abrije.tests.parser_fortest_err2)
        assert (
            abrije_.parsers == abrije.tests.parser_fortest_err2.expected_import_parser
        )

        # test import from list
        # make some dummy parsers for testing
        def rawparser_null(handle):
            return OrderedDict()

        def pathparser_none(path):
            return OrderedDict()

        expected_import_parser_list = {
            "null": ("rawparser_null", "raw", rawparser_null),
            "none": ("pathparser_none", "path", pathparser_none),
        }

        abrije_ = Abrije()  # reset test object
        toadd_list = [rawparser_null, pathparser_none]
        abrije_.import_parsers(toadd_list)
        assert abrije_.parsers == expected_import_parser_list

        # test import from tuple
        abrije_ = Abrije()  # reset test object
        toadd_tuple = (rawparser_null, pathparser_none)
        abrije_.import_parsers(toadd_tuple)
        assert abrije_.parsers == expected_import_parser_list

        # test import from dict
        abrije_ = Abrije()  # reset test object
        toadd_dict = {
            # use different names
            "parser_test1": rawparser_null,
            "pathparser_test2": pathparser_none,
        }

        expected_import_parser_dict = {
            "test1": ("parser_test1", "text", rawparser_null),
            "test2": ("pathparser_test2", "path", pathparser_none),
        }
        abrije_.import_parsers(toadd_dict)
        assert abrije_.parsers == expected_import_parser_dict

        # test import combined
        abrije_ = Abrije()  # reset test object
        expected_import_parser_combined = (
            abrije.tests.parser_fortest.expected_import_parser
        )
        expected_import_parser_combined.update(expected_import_parser_list)
        expected_import_parser_combined.update(expected_import_parser_dict)

        abrije_.import_parsers(abrije.tests.parser_fortest)
        abrije_.import_parsers(toadd_list)
        abrije_.import_parsers(toadd_dict)
        assert abrije_.parsers == expected_import_parser_combined

        # test re-import, causing warnings for same name
        with pytest.warns(UserWarning, match="same name"):
            abrije_.import_parsers(toadd_list)
        assert abrije_.parsers == expected_import_parser_combined



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abrije.parsers import *
import pprint

dedup_params = {
"parse_ignorecase" : "True",
"parse_unicode" : "True",
"parse_regex_reads1" : "[0-9\\-:, ]+ INFO Reads: .*(?P<heading>Input Reads): (?P<value>[0-9.]+)",
"parse_regex_reads2" : "[0-9\\-:, ]+ INFO Reads: .*(?P<heading>Read Pairs): (?P<value>[0-9.]+)",
"parse_regex_reads3" : "[0-9\\-:, ]+ INFO (?P<heading>.*reads.*out[^\\:]*): (?P<value>[0-9.]+)",
"parse_regex_position" : "[0-9\\-:, ]+ INFO (?P<heading>.+number.+position[^\\:]*): (?P<value>[0-9.]+)"
}

dedup_expected = (['Input Reads',
  'Read pairs',
  'Number of reads out',
  'Total number of positions deduplicated',
  'Mean number of unique UMIs per position',
  'Max. number of unique UMIs per position'],
 ['7318900', '7318900', '4270144', '3809138', '1.12', '235'])

dedup_input = "abrije/tests/dedup_example.log"

dedup_params1 = {
"heading_prefix" : "hpr_", # test adding prefix to heading (all)
"heading_suffix" : "_hsf", # test adding suffix to heading (all)
"value_prefix" : "vpr_", # test adding prefix to value (all)
"value_suffix" : "_vsf", # test adding suffix to value (all)
"parse_ignorecase" : "True",
"parse_unicode" : "True",
"parse_regex_reads1" : "[0-9\\-:, ]+ INFO Reads: .*(?P<heading1>Input Reads)(?P<heading2__rsf>): (?P<value1>[0-9.]+)(?P<value2__rsf>)",
    # above adds _rsf as suffix to heading and value
"parse_regex_reads2" : "[0-9\\-:, ]+ INFO Reads: .*(?P<heading1>Read Pairs)(?P<heading0_rpr_>): (?P<value1>[0-9.]+)(?P<value0_rpr_>)",
    # above adds rpr_ as prefix to heading and value
"parse_regex_reads3" : "[0-9\\-:, ]+ INFO (?P<heading1>.*reads.*out[^\\:]*): (?P<heading2>(?P<value>[0-9.]+))",
    # above appends value as suffix to heading
"parse_regex_position" : "[0-9\\-:, ]+ INFO (?P<heading>.+number.+position[^\\:]*): (?P<value0>(?P<value1>[0-9.]+))"
    # above appends value twice back to back
}

dedup_expected1 = (['hpr_Input Reads_rsf_hsf',
  'hpr_rpr_Read pairs_hsf',
  'hpr_Number of reads out4270144_hsf',
  'hpr_Total number of positions deduplicated_hsf',
  'hpr_Mean number of unique UMIs per position_hsf',
  'hpr_Max. number of unique UMIs per position_hsf'],
 ['vpr_7318900_rsf_vsf', 'vpr_rpr_7318900_vsf', 'vpr_4270144_vsf', 'vpr_38091383809138_vsf', 'vpr_1.121.12_vsf', 'vpr_235235_vsf'])

dedup_input1 = "abrije/tests/dedup_example.log"


rsem_params = {
    "parse_multiline" : "True",
    "parse_regex_N0_number_of_unalignable_reads" :
        "^([0-9.]+) [0-9.]+ [0-9.]+ [0-9.]+\n[0-9.]+ [0-9.]+ [0-9.]+\n[0-9.]+ [0-9.]+\n",
    "parse_regex_N1_number_of_alignable_reads" :
        "^[0-9.]+ ([0-9.]+) [0-9.]+ [0-9.]+\n[0-9.]+ [0-9.]+ [0-9.]+\n[0-9.]+ [0-9.]+\n",
    "parse_regex_N2_number_of_filtered_reads" :
        "^[0-9.]+ [0-9.]+ ([0-9.]+) [0-9.]+\n[0-9.]+ [0-9.]+ [0-9.]+\n[0-9.]+ [0-9.]+\n",
    "parse_regex_N_tot_total_number_of_reads" :
        "^[0-9.]+ [0-9.]+ [0-9.]+ ([0-9.]+)\n[0-9.]+ [0-9.]+ [0-9.]+\n[0-9.]+ [0-9.]+\n",

    "parse_regex_nUnique_number_of_reads_aligned_uniquely_to_a_gene" :
        "^[0-9.]+ [0-9.]+ [0-9.]+ [0-9.]+\n([0-9.]+) [0-9.]+ [0-9.]+\n[0-9.]+ [0-9.]+\n",
    "parse_regex_nMulti_number_of_reads_aligned_to_multiple_genes" :
        "^[0-9.]+ [0-9.]+ [0-9.]+ [0-9.]+\n[0-9.]+ ([0-9.]+) [0-9.]+\n[0-9.]+ [0-9.]+\n",
    "parse_regex_nUncertain_number_of_reads_aligned_to_multiple_locations" :
        "^[0-9.]+ [0-9.]+ [0-9.]+ [0-9.]+\n[0-9.]+ [0-9.]+ ([0-9.]+)\n[0-9.]+ [0-9.]+\n",

    "parse_regex_nHits_number_of_total_alignments" :
        "^[0-9.]+ [0-9.]+ [0-9.]+ [0-9.]+\n[0-9.]+ [0-9.]+ [0-9.]+\n([0-9.]+) [0-9.]+\n",
    "parse_regex_read_type" :
        "^[0-9.]+ [0-9.]+ [0-9.]+ [0-9.]+\n[0-9.]+ [0-9.]+ [0-9.]+\n[0-9.]+ ([0-9.]+)\n",
}

rsem_expected = (['N0_number_of_unalignable_reads',
  'N1_number_of_alignable_reads',
  'N2_number_of_filtered_reads',
  'N_tot_total_number_of_reads',
  'nUnique_number_of_reads_aligned_uniquely_to_a_gene',
  'nMulti_number_of_reads_aligned_to_multiple_genes',
  'nUncertain_number_of_reads_aligned_to_multiple_locations',
  'nHits_number_of_total_alignments',
  'read_type'],
 ['429826',
  '18765323',
  '0',
  '19195149',
  '17610329',
  '1154994',
  '14360103',
  '93404229',
  '3'])

rsem_input = "abrije/tests/rsem_example.cnt"

def test_regex_parser_dedup():
    with open(dedup_input) as inputfile:
        res = textparser_regex(inputfile, dedup_params)
        pprint.pprint(res) # print for updating expected, if needed
        assert res==dedup_expected

    with open(dedup_input1) as inputfile:
        res = textparser_regex(inputfile, dedup_params1)
        pprint.pprint(res) # print for updating expected, if needed
        assert res==dedup_expected1


def test_regex_parser_rsem():
    with open(rsem_input) as inputfile:
        res = textparser_regex(inputfile, rsem_params)
        pprint.pprint(res)
        assert res==rsem_expected

# for printing expected
def main():
    test_regex_parser_dedup()
    test_regex_parser_rsem()

if __name__ == '__main__':
    main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# uses Google docstrings: https://google.github.io/styleguide/pyguide.html
# (numpy-style not supported in multi-line returns docstrings with type hints)
# use typehints, not in docstrings

"""
abrije
======

abrije is a generic log parser and summariser.

A user provides ``abrije`` with a list of `logset`s to parse. Each `logset`
consists of name, a snakemake-style path pattern, a type to specify method
of parsing the log file, and parameters to go with the parser. A general
config can also be provided to give defaults for all `logset`s. 

Old (out-of-date) docs follow:

To process input files, ``abrije`` will test each regex sequential until it
finds one that matches the input file path (as determined by the glob). Named
groups with the regex provide required or optional info to ``abrije``. All
named groups should be in the format ``prefixSUFFIX__providedcontents`` or
``prefix_SUFFIX__providedcontents``. SUFFIX and providedcontents are usually
optional. Unless otherwise noted, providedcontents simply replaces the contents captured by the
regex to provide additional flexibility.
The
following prefixes are possible, each serving a different purpose:

    1) label : This prefix is used for given the input file a label. Multiple
    levels of label are possible, these will be sorted by label name before
    the contents (i.e. text captured by the regex or providedcontents) is combined into a 'full_label' separated by '>' characters. Since the
    label names are sorted, it is useful to provide numbered suffixes to these
    (e.g. label_1run, label2_sample to give Run129>Sample141) so the ordering is clearly specified.
    Data from all input files that have the same full_label are combined into
    a single output row. At least one label is required.

    2) type : This prefix is used to specify the file type for the input file
    which determines the parser used to process the file. A parser is simply
    a Python function that takes a file handle as single parameter, and returns
    an OrderedDict of data column_heading:value. If a suffix is provided, it is used
    as the file type, otherwise the captured contents of the regex is used. The
    providedcontents does not override the capture contents, and instead provides
    additional input settings in the format setting1_IS_value1_AND_setting2_IS_value2.
    Currently, the only input setting available is 'encoding' which specifies the
    input encoding of the file (overriding the default script option). An additional
    'encoding' provided by ``abrije`` is 'binary' which causes the input file
    to be opened in binary mode, instead of text mode.

    3) tag : tags are used to 'tag' the columns read from the input file. It is
    useful to use this when multiple files of the same type are present for the
    same full_label (e.g. different steps). Tagging allows the data from each
    file to be given a unique column heading. Each tag is in the added in the format
    SUFFIX=contents, or simply the SUFFIX or contents if the other is absent. Multiple
    tags are separated by commas, and all tags are prepended to the original column heading
    followed by a '>'.

    4) comment : comment specifies additional columns prepended to any input from
    the data file. The column heading is SUFFIX (or 'comment' if absent), and the contents
    are the value. All comment columns are also tagged.

    ``abrije`` locates the function as
    type_parse() in predefined parsers or a module that is imported using the


Working from a directory tree, the
A list of globs and regexes

list of globs. Used to locate input files.

Each glob is used by ``abrije`` to build a list of input files to parse.

list of regexes. Used to provide metadata and tell abrije how to parse the file.
The path of each input file is tested against each regex for a match. Upon
finding a match (i.e. the first matching regex is used), the named groups
within the regex are used as follows:
label, labelXXXX, or label_XXXX = used for grouping log files
type, typeXXXX, or type_XXXX = used for specifying log type, former group content gives type, otherwise use XXXX
tag, tagXXXX, or tag_XXXX = used to tag log data in heading, former tags with 'tagged', otherwise tag with XXXX
comment, commentXXXX or comment_XXXX
first underscore before XXXX ignored
labelXXXX__YYYY YYYY used as content instead of group_contents

provide ``abrije`` with key data about the file
# provide list of regex
#  how to match regex to file? simplest would be to use first regex that matches
# regex has named groups
#  label or label_x -> used to match data to a line
#  type or type_xxx
#    if group name is type, group match gives type_name
#    otherwise, group name gives is type_name(xxx); actual match ignored, allows regex definition to define type, as filename may not have discernable type
#    used to define the parser, will call type_parse(file)
#    expected to return a dict containing headings>data
#  tag_xxxx
#   tags headings with xxxx (group match value ignored)
#  other groups tag headings with groupname=value
#

 path of each input file is tested against each regex for a match.


plugins can use logger 'abrije'
"""


###
# smart parse files
# provide list of files
# provide list of regex
#  how to match regex to file? simplest would be to use first regex that matches
# regex has named groups
#  label or label_x -> used to match data to a line
#  type or type_xxx
#    if group name is type, group match gives type_name
#    otherwise, group name gives is type_name(xxx); actual match ignored, allows regex definition to define type, as filename may not have discernable type
#    used to define the parser, will call type_parse(file)
#    expected to return a dict containing headings>data
#  tag_xxxx
#   tags headings with xxxx (group match value ignored)
#  other groups tag headings with groupname=value
#

import csv
import re
import pathlib
import glob
import importlib
import argparse
import sys
import logging
import pprint
import traceback
import os
import warnings
import typing
import json
import yaml

from collections import OrderedDict, defaultdict, namedtuple
from types import ModuleType
from typing import (
    Callable,
    Dict,
    IO,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

try:  # workaround for Python<3.7, no typing.OrderedDict
    typing.OrderedDict
except AttributeError:
    typing.OrderedDict = typing.MutableMapping


import snakemake.utils 
try:
    import abrije.parsers
except:
    # this allows importing abrije.parsers even when run as a script
    import parsers
    class Namespace:
        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)
    abrije = Namespace(parsers=parsers)

import io


_PROGRAM_NAME = "abrije"
_PROGRAM_VERSION = "0.6.0.dev2"


# INIT LOGGING
# logging, create a stderr and stdout handler
logger = logging.getLogger(_PROGRAM_NAME)  #: Module logger.
logger.setLevel(logging.DEBUG)  # top-level logs everything

stderr_handler = logging.StreamHandler(sys.stderr)
"""
Logging handler to `stderr`, `WARNING` and above.
"""
logger.addHandler(stderr_handler)
stderr_formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
stderr_handler.setFormatter(stderr_formatter)
stderr_handler.setLevel(logging.WARNING)

stdout_handler = logging.StreamHandler(sys.stdout)
"""
Logging handler to `stdout`, controlled by verbosity.
"""
logger.addHandler(stdout_handler)
stdout_formatter = logging.Formatter(
    "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
stdout_handler.setFormatter(stdout_formatter)

# warning handling
def warning(msg: str) -> None:
    logger.warning(msg)
    warnings.warn(msg)


# integration between logging and warnings not used as this logs to separate
# logger
# warnings.simplefilter("always")
# logging.captureWarnings(True)  # integrate with warnings.warn
# py_warnings_logger = logging.getLogger("py.warnings")
# py_warnings_logger.addHandler(stdout_handler)
# py_warnings_logger.addHandler(stderr_handler)

_DEBUG: bool = False  #: `True` enables traceback.
_verbose: int = 0  #: Controls output to `stdout` handler.


def set_verbosity(verbose: Optional[int]) -> None:
    """
    Set logging verbosity to stdout:
    `2` : `DEBUG` and enables traceback.
    `1` : `INFO`.
    `0` : `CRITICAL`.
    """
    global _verbose, _DEBUG
    if verbose is None:
        verbose = 0
    _verbose = verbose
    _DEBUG = False
    if _verbose >= 2:
        stdout_handler.setLevel(logging.DEBUG)
        _DEBUG = True
        # if verbosity is debug, traceback of handled exceptions will be output as logger.error
    elif _verbose >= 1:
        stdout_handler.setLevel(logging.INFO)
    elif _verbose >= 0:
        stdout_handler.setLevel(logging.CRITICAL)


set_verbosity(2)

# pretty formatter, wrap in object so not called unless needed
pformat = pprint.PrettyPrinter(indent=2, width=120).pformat


class PrettyLog:
    def __init__(self, obj):
        self.obj = obj

    def __repr__(self):
        # if isinstance(self.obj,OrderedDict):
        # special handling for OrderedDict, as not handled by pformat by default
        #    return pformat([(key,value) for (key,value) in self.obj.items()])
        # pprint SEEMS to work in Python 3.6
        return pformat(self.obj)


ParserFunc = Callable[
    [Union[IO[str], IO[bytes], str], Dict[str, str]], Tuple[List[str], List[str]]
]
"""
Parsers are in the form `parser(handle_or_path, tokens)` where `handle_or_path`
is the input file handle or path and `tokens` is a `dict` of tokens for the
log file. Parsers should return a tuple `(headings, values)` where `headings`
and `values` are lists.
"""

StoredParser = Tuple[str, str, ParserFunc]
"""
Parser details stored as `(fullname, type, function)`.
"""

TNamedTokens = TypeVar("TNamedTokens", bound="NamedTokens")


class NamedTokens:
    """
    A simple class that maintains parallel lists of `headings` and `values`.

    Ideally created from factory methods `new()` and `fromLists()`.
    """

    def __init__(self, headings: List[str], values: List[str]) -> None:
        self.headings: List[str] = headings  #: List of `headings`.
        self.values: List[str] = values  #: List of `values`.

    @classmethod
    def new(cls, heading: str = None, value: str = None) -> TNamedTokens:
        """
        Create a new `NamedTokens` object, with optional first values; object
        starts empty otherwise.

        Args:
            heading: First `headings` value.
            value: First `values` value.

        Returns:
            A new `NamedTokens` object.
        """
        obj = NamedTokens([], [])
        if heading is not None:
            assert value is not None, "both heading and value should be provided"
            obj.append(heading, value)
        return obj

    @classmethod
    def fromLists(cls, headings: List[str], values: List[str]) -> TNamedTokens:
        """
        Create a new `NamedTokens` object from the provided `headings` and 
        `values` lists. The provided lists are used directly by the new object
        and may be mutated by the object; they should only be mutated externally
        with care to avoid corrupting the `NamedTokens` object using them.

        Args:
            headings: List of `headings` for this object.
            values: List of `values` for this object. Should be same length as
                `headings`.

        Returns:
            A new `NamedTokens` object.
        """
        assert len(headings) == len(
            values
        ), "headings and values should be the same size"
        return NamedTokens(headings, values)

    def append(self, heading: str, value: str) -> None:
        """
        Append `heading` and `value` to this object.

        Args:
            heading: The `headings` value to append.
            value: The `values` value to append.
        """
        self.headings.append(heading)
        self.values.append(value)

    def extend(self, items: Iterable[Tuple[str, str]]) -> None:
        """
        Extend this object with additional `headings` and `values`.

        Args:
            items: An iterable yielding tuples of `(heading, value)` to add to
                this object.
        """
        for heading, value in items:
            self.headings.append(heading)
            self.values.append(value)

    def tag_headings(self, prefix: str) -> None:
        """
        Tag headings by prepending `prefix` to each.

        Args:
            prefix: Prefix to add to each heading.
        """
        for i, heading in enumerate(self.headings):
            self.headings[i] = prefix + heading

    def render_values(self) -> str:
        """
        Convert contents to `heading1:value1, heading2:value2, ...`.

        Returns:
            `heading1:value1, heading2:value2, ...`.
        """
        return ", ".join(f"{heading}:{value}" for heading, value in self)

    def has_redundant_headings(self) -> bool:
        """
        Check if any `headings` are identical.

        Returns:
            `True` if there are any identical headings, `False` otherwise.
        """
        return len(self.headings) > len(set(self.headings))

    def make_headings_unique(self) -> None:
        """
        Makes all headings unique by adding `(x)` as suffix where `x` is an 
        increasing number from 2 until a unique heading is found.
        """
        existing = {}

        for i, heading in enumerate(self.headings):
            suffix_index = 2
            newheading = heading
            while newheading in existing:
                newheading = f"{heading} ({suffix_index})"
                suffix_index += 1
            self.headings[i] = newheading
            existing[newheading] = True

    def __len__(self) -> int:
        return len(self.headings)

    def __iter__(self) -> Iterator[Tuple[str, str]]:
        return zip(self.headings, self.values)

    def __repr__(self) -> str:
        return (
            "{" + ", ".join(f"'{heading}':'{value}'" for heading, value in self) + "}"
        )

    def __eq__(self, o: object) -> bool:
        if isinstance(o, NamedTokens):
            return self.headings == o.headings and self.values == o.values
        else:
            return NotImplemented

# TODO: implicitly expects tokens to be an OrderedDict. Is this guaranteed? Perhaps in python 3.7+
# (Implementation detail in python 3.6)
def filter_tokens(source: TNamedTokens, tokens: Dict[str, str]) -> TNamedTokens:
    _istrue = lambda tokens, name: name in tokens and yaml.safe_load(tokens[name]) == True
    _getvalue = lambda tokens, name, default = None: tokens[name] if name in tokens else default

    re_flags = 0
    ignore_case = _istrue(tokens, "filter_ignorecase")
    use_unicode = _istrue(tokens, "filter_unicode")
    filter_search = _istrue(tokens, "filter_search")
    filter_default_exclude = _istrue(tokens, "filter_default_exclude")

    if ignore_case: re_flags |= re.IGNORECASE
    if use_unicode: re_flags |= re.UNICODE

    regex_param_re = re.compile("^filter_(exclude|include).*$")

    re_list = []
    for token, value in tokens.items():
        token_match = regex_param_re.match(token)
        if token_match:
            filter_exclude = token_match.group(1)=="exclude"
            re_compiled = re.compile(value, re_flags)
            re_list.append((filter_exclude, re_compiled))

    print(re_list)
    obj = NamedTokens.new()
    for heading, value in source:
        exclude = filter_default_exclude
        for filter_exclude, re_compiled in re_list:
            if filter_search:
                re_match = re_compiled.search(heading)
            else:
                re_match = re_compiled.match(heading)
            if re_match:
                exclude = filter_exclude
        if not exclude: obj.append(heading, value)
    return obj

class AbrijeConfig:
    """
    Class for storing config used when processing `log_set`s with `abrije`.

    A separate class is used to facilitate a nested defaults (for all `log_sets`)/
    override config for a particular `log_set`.

    Args:
    config_set: Name of this config set.

    """

    wildcard_param_regex = re.compile("^{(.+)}$")
    """
    Wildcard parameters are specified as `{wildcard_name}`.
    """

    def __init__(self, config_set: str):
        self.config_set: str = config_set
        """
        Name of this config set.
        """

        self.wildcard_mapping: Optional[Dict[str, str]] = None
        """
        Stores mapping of `wildcard_name -> token_name` where `token_name` is
        a `label_name` or `tag_name` identifier used to determine what the
        token value (wildcard contents) should be used for by `abrije`,
        e.g. `'sample' -> 'label1_samplename'`.
        """

        self.const_mapping: Optional[Dict[str, str]] = None
        """
        Stores mapping of `token_name -> token_value` to give a constant 
        `token_value` for the given `token_name`. Constants have precedence
        over wildcards with the same `token_name`.
        e.g. `'label1_samplename' -> 'Sample140'`
        """

        self.overrides: Optional[Dict[str, AbrijeConfig]] = None
        """
        Stores mapping of `log_set -> override_config` where 
        `override_config` is the (optional) config for a specific
        `log_set`.
        """

    def add_to_config(
        self, parameters=Dict[str, str], log_set: Optional[str] = None
    ) -> None:
        """
        Adds `parameters` to this object (if `log_set` is `None`) or to the
        override config for given `log_set`.

        Args:
            parameters: `dict` of `param_name -> param_value`.  Wildcard 
                parameters are specified as `{wildcard_name}` in the 
                `param_name`, all others (without curly brackets) are 
                considered const parameters.
            log_set: Name of `log_set` if these are for override config of that 
                `log_set`.
        """

        if log_set == "defaults":
            raise ValueError("add_to_config: log_set = defaults should not be used")
        assert (
            log_set != self.config_set
        ), "add_to_config: log_set should not be passed to overrides"

        if log_set is None:
            for param_name, param_value in parameters.items():
                wildcard_param_match = self.wildcard_param_regex.match(param_name)
                if wildcard_param_match:
                    # wildcard param
                    if not self.wildcard_mapping:
                        self.wildcard_mapping = {}
                    wildcard_param_name = wildcard_param_match.group(1)
                    self.wildcard_mapping[wildcard_param_name] = param_value
                    logger.debug(
                        f"Added wildcard parameter {wildcard_param_name}:{param_value} "
                        f"to {self.config_set}"
                    )
                else:
                    # const param
                    if not self.const_mapping:
                        self.const_mapping = {}
                    self.const_mapping[param_name] = param_value
                    logger.debug(
                        f"Added const parameter {param_name}:{param_value} "
                        f"to {self.config_set}"
                    )

        else:
            if not self.overrides:
                self.overrides = {}
            abrije_config = self.overrides.setdefault(log_set, AbrijeConfig(log_set))
            abrije_config.add_to_config(parameters)

    def resolve_tokens(
        self,
        log_set: Optional[str],
        wildcards: Optional[Dict[str, str]],
        cur_tokens: Optional[Dict[str, str]] = None,
        consumed_wildcards: Optional[Set[str]] = None,
    ) -> Tuple[Dict[str, str], Optional[Set[str]]]:
        """
        Resolve all wildcard or const parameter tokens.

        Args:
            log_set: Name of the `log_set` to resolve for; used to detect any 
                overrides. May be `None`, in which case no overrides are 
                resolved.

            wildcards: Wildcards as `dict` of `wildcard_name -> wildcard_value` 
                used to resolve any wildcard parameters. May be `None`, in which 
                case only const tokens are resolved.

            cur_tokens: A `dict` of any tokens already resolve; new tokens are 
                added to this if provided, replacing any existing tokens of the 
                same name.

            consumed_wildcards: A `set` of any wildcards already consumed; if
                provided, any additional wildcards that are consumed when 
                resolving are added to this.


        Returns:
            `(param_dict, consumed_wildcards_set)` where param_dict is a `dict`
            of `param_name -> param_value` and `consumed_wildcards_set` is a
            `set` of all wildcards that were consumed when resolving tokens.
            These will be the same as the parameters `cur_tokens` and
            `consumed_wildcards` if provided. `consumed_wildcards_set` will be 
            `None` if `wildcards` was `None`.

        """
        if cur_tokens is None:
            cur_tokens = {}

        if wildcards is None:
            consumed_wildcards = None
        else:
            if consumed_wildcards is None:
                consumed_wildcards = set()

            if self.wildcard_mapping:
                for wildcard_name, wildcard_value in wildcards.items():
                    if wildcard_name in self.wildcard_mapping:
                        token_name = self.wildcard_mapping[wildcard_name]
                        cur_tokens[token_name] = wildcard_value
                        consumed_wildcards.add(wildcard_name)
                        logger.debug(
                            f"Added token {token_name}:{wildcard_value} "
                            f"from wildcard {wildcard_name} for {self.config_set}"
                        )

        if self.const_mapping:
            for token_name, token_value in self.const_mapping.items():
                cur_tokens[token_name] = token_value
                logger.debug(
                    f"Added token {token_name}:{token_value} " f"for {self.config_set}"
                )

        if log_set is not None and self.overrides and log_set in self.overrides:
            self.overrides[log_set].resolve_tokens(
                log_set, wildcards, cur_tokens, consumed_wildcards
            )  # cur_tokens and consumed_wildcards modified in-place

        return cur_tokens, consumed_wildcards


class Abrije:
    """
    Main class for `abrije`.


    """

    def __init__(self):
        self.parsers: Dict[str, StoredParser] = {}
        """
        Stores all loaded parsers, maps `name1 -> (fullname, type, function)`
        e.g. `'csv_firstrow' -> ('parser_csv_firstrow', 'text', parser_csv_firstrow())`.
        """

        self.log_sets: typing.OrderedDict[str, str] = OrderedDict()
        """
        Stores input patterns to search for log files to process,
        as `log_set_name -> log_set_pattern` with the `log_set_pattern` in 
        snakemake wildcard syntax
        e.g. `'rule_align' -> '{step}/{sample}-{genome}.sam'`.
        """

        self.parser_mapping: Dict[str, str] = {}
        """
        Stores mapping of `log_set -> parser_name` where `parser_name` is
        the name of a parser used for processing the `log_set`.
        """

        self.default_config = AbrijeConfig("defaults")
        """
        Default config for all `log_sets`; parent for all override configs.
        """

        self.all_headings: typing.OrderedDict[str, str] = OrderedDict()
        """
        Used to track heading order.
        Stores a set of all headings in order seen by the algorithm as
        `heading -> heading_type` where `heading_type` is a string later used 
        to reorder headings according to type (e.g. labels first).
        """

        LogRow = Dict[str, str]
        """
        A row in the output stored `heading -> value`.
        """

        self.output_rows: typing.Dict[str, LogRow] = {}
        """
        Stores rows of log data read from `log_set`s as `full_label -> row`; 
        does not need to be ordered as `output_rows` sorted by `full_label` 
        before output generated.
        """

    def add_log_set(self, log_set_name: str, log_set_pattern: str, log_set_type: str):
        """
        Add a `log_set`.

        Args:
            log_set_name: Name of the `log_set`.
            log_set_pattern: Snakemake-style wildcard pattern used to find log
                files to parse for this `log_set`.
            log_set_type: 
        """
        logger.debug(
            f"Adding log_set: "
            f"name {log_set_name}, pattern {log_set_pattern}, type {log_set_type}"
        )
        if log_set_name == "" or log_set_pattern == "" or log_set_type == "":
            raise ValueError("log_set name, pattern and type must not be blank")
        self.log_sets[log_set_name] = log_set_pattern
        self.parser_mapping[log_set_name] = log_set_type

    def add_parser(self, fullname: str, func: ParserFunc) -> None:
        """
        Helper to add a parser to stored parsers, with error and warning
        handling for incompatible types or attempts to replace parsers.
        Only parsers with name starting with the parser prefix 
        `([a-z]*)parser_` are imported, others are silently ignored.
        The prefix before parser determines the parser type. 
        Allowed parser types are `text`, `svtext`, `raw`, and `path` 
        If no type is specified (`parser_`), the default is `text`.
        The text following the prefix gives the parser name. 
        (e.g. `pathparser_test` is a parser named `test` of type `path`).

        Readding a parser of the same name (regardless of type) produces a  
        warning, and the new parser is *not* added.

        Args:
            fullname: The full name (including the parser type prefix) of the
                parser e.g. `pathparser_test`.
            func: The parser function itself.

        Raises:
            ValueError: If `func` is not callable or an invalid parser type 
                specified.
        """
        match = re.match("^([a-z]*)parser_(.*)$", fullname)
        if match:
            parser_type = match.group(1)
            if parser_type == "":
                parser_type = "text"  # default
            parser_name = match.group(2)
            if not callable(func):
                raise ValueError(f"cannot import {fullname} as parser - not callable")
            if not parser_type in ["text", "svtext", "raw", "path"]:
                raise ValueError(
                    f"cannot import {fullname} as parser - unknown type {parser_type}"
                )
            if parser_name in self.parsers:
                warning(
                    f"ignoring parser {fullname} - parser of same name already exists"
                )
                return
            logger.debug(
                f"Adding parser: name {parser_name}"
                f"fullname {fullname}, parser_type {parser_type}, func {func}"
            )
            self.parsers[parser_name] = (fullname, parser_type, func)
        else:
            logger.debug(f"Ignoring (not a parser): {fullname}")

    def import_parsers(
        self,
        import_from: Union[
            ModuleType, Tuple[ParserFunc, ...], List[ParserFunc], Dict[str, ParserFunc]
        ],
    ) -> None:
        """
        Import parsers from a module, `tuple`, `list` or `dict`. 

        A parser is a function that takes parses a log file. 
        Parsers are identified by name.
        The `fullname` of a parser is: `{prefix}parser_{name}`. The `prefix`
        is the type of the parser, while `name` identifies the parser for
        downstream use.

        The types of parser are (given by `prefix`):
        - `text` (default): parameter is open text file handle to parse
        - `svtext`: parameter is open text file with `newline=''`, meant to
          be fed into Python `csv.read()`
        - `raw`: parameter is open binary file handle to parse
        - `path`: parameter is path to a file to parse

        A blank prefix `''` indicates the default parser type.

        A parser function takes (first arg) an open file handle of a text (`text` 
        parser), text with `newline='' (`svtext` ) or binary (`raw`) file 
        (for reading), or a file path string (`path`), (second arg) a dictionary 
        of tokens (potentially containing parse parameters), and returns 
        parallel lists of headings and values as a tuple 
        i.e. `([heading1, heading2, ...], [value1, value2, ...])`.
        
        If `import_from` is a module, all parsers (identified by a valid
        `fullname`) are important from the module.

        If `import_from` is a list or tuple, the `fullname` is obtained from
        the function name itself.

        If `input_from` is a the dict, the `fullname` is the key and the value 
        is the function.

        Args:
            import_from: Module, tuple, list or dict of parsers to import.

        Raises:
            ValueError: If `import_from` as unsupported type, a parser is invalid 
            type or a parser function is not callable.
        """

        if isinstance(import_from, ModuleType):
            logger.debug(f"Adding parsers from module: {import_from.__name__}")
            for name in dir(import_from):
                self.add_parser(name, getattr(import_from, name))
        elif isinstance(import_from, tuple) or isinstance(import_from, list):
            logger.debug(f"Adding parsers from: {import_from}")
            for item in import_from:
                try:
                    name = item.__name__
                except:
                    raise ValueError(
                        f"cannot import {item} as parser - no __name__"
                    ) from None
                self.add_parser(name, item)
        elif isinstance(import_from, dict):
            logger.debug(f"Adding parsers from: {import_from}")
            for name, item in import_from.items():
                self.add_parser(name, item)
        else:
            raise ValueError(f"cannot import from unsupported type {type(import_from)}")

    def add_to_config(
        self, parameters: Dict[str, str], log_set: Optional[str] = None
    ) -> None:
        """
        Adds `parameters` as defaults (if `log_set` is `None` or `"defaults"`) 
        or to the override config for given `log_set`.

        Args:
            parameters: `dict` of `param_name -> param_value`.  Wildcard 
                parameters are specified as `{wildcard_name}` in the 
                `param_name`, all others are considered const parameters.
            log_set: Name of `log_set` if these are for override config of that 
                `log_set`.
        """
        if log_set == "defaults":
            # "defaults" is not a log_set, strip value
            log_set = None

        self.default_config.add_to_config(parameters, log_set)
        # note: no check if log_set actually exists as cannot guarantee log_sets
        # added before config, check_config() does this before processing

    def check_config(self):
        """
        Performs a sanity check to ensure override parameter names match `log_set`s.
        Meant to catch potential errors caused by misspelt `log_set`s/overrides.
        """
        if len(self.log_sets) == 0:
            raise ValueError("no log_sets to process")
        if self.default_config.overrides is None or set(self.log_sets) >= set(
            self.default_config.overrides.keys()
        ):
            pass
        else:
            raise ValueError("some overrides do not match log_set names")

    def resolve_tokens(
        self, log_set: Optional[str], wildcards: Optional[Dict[str, str]]
    ) -> Dict[str, str]:
        """
        Resolve all wildcard or const parameter tokens for a `log_set`.

        Args:
            log_set: Name of the `log_set` to resolve for; used to detect any 
                overrides. May be `None`, in which case no overrides are 
                resolved.

            wildcards: Wildcards as `dict` of `wildcard_name -> wildcard_value` 
                used to resolve any wildcard parameters. May be `None`, in which 
                case only const tokens are resolved.

        Raises:
            RuntimeError: If there were any wildcards not converted to tokens.

        Returns:
            A `dict` of `param_name -> param_value`.
        """
        tokens, consumed_wildcards = self.default_config.resolve_tokens(
            log_set, wildcards
        )
        if wildcards is not None:
            assert (
                consumed_wildcards is not None
            ), "some wildcards consumed when none provided?"
            assert consumed_wildcards <= set(
                wildcards  # this shouldn't occur, can't consume any wildcards not provided
            ), f"consumed unknown wildcards {consumed_wildcards - set(wildcards)}"
            unconsumed_wildcards = set(wildcards) - consumed_wildcards
            if unconsumed_wildcards:
                raise RuntimeError(
                    f"the following wildcards were not converted to tokens: {unconsumed_wildcards}"
                )
            logger.info(f"Resolved tokens for log_set {log_set}: {tokens}")
        else:
            logger.info(f"Resolved universal tokens: {tokens}")
        return tokens

    def process_log_sets(self) -> None:
        """
        Processes all log files given in `log_set`s. Processed output stored in
        object.

        Raises:
            RuntimeError: If this function has already been called.
        """

        self.check_config()

        if self.all_headings:
            raise RuntimeError("second call to process_log_sets() - not supported")

        # logic pseudocode
        # for each log_set_name, log_pattern in self.log_sets:
        #     list files for pattern
        #     for file in file list:
        #         have wildcard values -> get labels, tags, comments
        #             either from wildcards or const, with overrides
        #             logic:
        #                 1) find any const
        #                 2) find any wildcards
        #                 3) if wildcard does not exist, no token
        #         get parser -> process file -> get output dict
        #         make full_label from labels
        #         add comments
        #         tag headings in output
        #         group with any existing full_label, combine headings and data
        # sort full output by full_label
        # write
        logger.info("Processing log sets...")
        # iterate log_sets
        for log_set_name, log_set_pattern in self.log_sets.items():
            parser = self.get_parser(log_set_name)
            logger.debug(
                f"Listing files for log_set {log_set_name} with "
                f"pattern {log_set_pattern}"
            )

            logfile_n = 0
            for logfile_n, (logpath, wildcards_list) in enumerate(
                snakemake.utils.listfiles(log_set_pattern)
            ):
                wildcards = dict(wildcards_list.items())

                logger.debug(
                    f"Processing log file: path {logpath}, wildcards {wildcards}"
                )
                tokens = self.resolve_tokens(log_set_name, wildcards)
                sorted_token_keys = sorted(
                    tokens.keys()
                )  # ensure we sort keys to get same output regardless of parameter order

                def get_tokens_with_prefix(prefix):
                    for key in sorted_token_keys:
                        if key.startswith(prefix):
                            yield (key, tokens[key])

                labels = NamedTokens.new()
                labels.extend(get_tokens_with_prefix("label"))
                comments = NamedTokens.new()
                comments.extend(get_tokens_with_prefix("comment"))
                tags = NamedTokens.new("log_set", log_set_name)
                # log_set is always the first tag
                tags.extend(get_tokens_with_prefix("tag"))

                if len(labels) == 0:
                    logger.info(f"No label tokens found, using path as label")
                    labels.append("labelpath", logpath)

                logger.debug(f"Labels: {labels}")
                logger.debug(f"Comments: {comments}")
                logger.debug(f"Tags: {tags}")

                full_label = ">".join(labels.values)
                logger.debug(f"Full label from is {full_label}")

                log_raw_output = NamedTokens.fromLists(
                    *self.parse_path(parser, logpath, tokens)
                )
                logger.debug(f"Parser output is {log_raw_output}")

                log_output = filter_tokens(log_raw_output, tokens)
                logger.debug(f"Filtered output is {log_raw_output}")

                if len(log_output) == 0:
                    warning("output is empty")

                if log_output.has_redundant_headings():
                    logger.info("Output has redundant headings, making all unique")
                    log_output.make_headings_unique()

                tag_prefix = tags.render_values() + " "
                log_output.tag_headings(tag_prefix)
                comments.tag_headings(tag_prefix)

                logger.debug(f"Tagged comments headings: {comments.headings}")
                logger.debug(f"Tagged headings: {log_output.headings}")

                output_row = self.output_rows.setdefault(full_label, {})

                def add_to_output(heading, value):
                    if heading not in self.all_headings:
                        self.all_headings[heading] = True
                    if heading in output_row and output_row[heading] != value:
                        warning(
                            f"replacing value {heading} for {full_label}, "
                            "add a tag to prevent redundant headings"
                        )
                        # should virtually never happen as log_set, which must be
                        # unique, is used as tag

                    output_row[heading] = value

                def add_all_to_output(named_tokens):
                    for heading, value in named_tokens:
                        add_to_output(heading, value)

                add_to_output("full_label", full_label)
                add_all_to_output(labels)
                add_all_to_output(comments)
                add_all_to_output(log_output)
            logger.info(f"Processed {logfile_n} files for log_set {log_set_name}.")
            if logfile_n == 0:
                warning(f"no log files processed for log_set {log_set_name}")

    # TODO: tests for no parser mapping or parser
    def get_parser(self, log_set_name: str) -> StoredParser:
        """
        Returns stored parser for given `log_set`.

        Args:
            log_set_name: `log_set` to get parser for.

        Raises:
            RuntimeError: If no parser set for the `log_set` or parser could not be 
            found.

        Returns:
            Stored parser for given `log_set`.
        """
        if log_set_name not in self.parser_mapping:
            raise RuntimeError(f"no parser set for log_set {log_set_name}")
        parser_name = self.parser_mapping[log_set_name]
        if parser_name not in self.parsers:
            raise RuntimeError(f"no parser {parser_name} for {log_set_name}")
        parser = self.parsers[parser_name]
        logger.debug(f"Using parser {parser_name} {parser} for log_set {log_set_name}")
        return parser

    # TODO: document open_encoding
    def parse_path(self, parser: StoredParser, path: str, tokens: Dict[str, str]):
        """
        Executes parser function for given path and tokens.

        Args:
            parser: Stored parcel to use.
            path: Log file path to parse.
            tokens: Tokens to pass to parse function (potentially containing 
                parse parameters).

        Returns:
            Output from parser function.
        """
        parser_fullname, parser_type, parser_function = parser
        input_encoding = tokens.get("open_encoding", "utf8")
        if parser_type == "text":
            logger.debug(
                f"'text' parser: Opening text log file {path} with encoding {input_encoding}"
            )
            with open(path, "r", encoding=input_encoding) as handle:
                output = parser_function(handle, tokens)
        elif parser_type == "svtext":
            logger.debug(
                f"'svtext' parser: Opening text log file {path} with encoding {input_encoding}, "
                f"newline ''"
            )
            with open(path, "r", newline="", encoding=input_encoding) as handle:
                output = parser_function(handle, tokens)
        elif parser_type == "raw":
            logger.debug(f"'raw' parser: Opening binary log file {path}")
            with open(path, "rb") as handle:
                output = parser_function(handle, tokens)
        elif parser_type == "path":
            logger.debug(f"'path' parser: Providing path {path} to parser")
            output = parser_function(path, tokens)
        else:
            raise AssertionError(f"unexpected parser_type {parser_type}")
        return output

    # TODO: Ensure we write lables first, and in order?
    def write_output(self, outfile: Union[IO[str], str]) -> None:
        """
        Write output to a open file handle or path.

        Args:
            outfile: An open file handle or path string.

        Raises:
            RuntimeError: If called before processing complete.
        """
        if not self.all_headings:
            raise RuntimeError("write_output() called before process_log_sets()")

        tokens = self.resolve_tokens(None, None)

        output_encoding = tokens.get("output_encoding", "utf8")
        output_navalue = tokens.get("output_navalue", "")
        output_delimiter = tokens.get("output_delimiter", "\t")

        if isinstance(outfile, str):
            csv_file = open(outfile, "w", newline="", encoding=output_encoding)
            logger.info(
                f"Opening output file {outfile} with encoding {output_encoding}"
            )
        else:
            csv_file = outfile

        csv_writer = csv.DictWriter(
            csv_file,
            restval=output_navalue,
            delimiter=output_delimiter,
            fieldnames=self.all_headings,
        )

        logger.debug(
            f"Output delimiter {output_delimiter!r}, NA value {output_navalue!r}"
        )

        logger.debug(f"Writing header, total {len(self.all_headings)} columns")
        csv_writer.writeheader()
        for full_label in sorted(self.output_rows.keys()):
            row = self.output_rows[full_label]
            logger.debug(f"Writing row {full_label}, total {len(row)} valid values")
            csv_writer.writerow(row)

        if csv_file is not outfile:
            csv_file.close()


# TODO move test to unit test
def test():
    abrije_ = Abrije()
    abrije_.import_parsers(abrije.parsers)
    abrije_.add_log_set(
        "step0",
        "logs/0_subsampled_data/{sample_name}___{inputid}_{end_label}.fastq.stats",
        "tsv_wide",
    )

    abrije_.add_log_set(
        "step1",
        "logs/1_trimmed/{sample_name}___{inputid}_{end_label}.fastq.stats",
        "tsv_wide",
    )
    abrije_.add_to_config(
        {
            "{sample_name}": "label1_sample_name",
            "{end_label}": "label3_end_label",
            "{inputid}": "label2_inputid",
            "open_encoding": "utf8",
            "tag2_best": "bag1",
            # "output_delimiter": ",",
        },
    )
    abrije_.add_to_config(
        {"tag1_test": "tag1", "comment1_det": "234"}, log_set="step0",
    )
    abrije_.process_log_sets()
    abrije_.write_output("test.tsv")

    import shlex

    config = {
        "log_set": [
            (
                "step1",
                "logs/1_trimmed/{sample_name}___{inputid}_{end_label}.fastq.stats",
                "tsv_wide",  # allow optional dict as config
            ),
            (
                "step1",
                "logs/1_trimmed/{sample_name}___{inputid}_{end_label}.fastq.stats",
                "tsv_wide",
            ),
        ],  # allow top level defaults dict
        "config": {
            "defaults": {
                "{sample_name}": "label1_sample_name",
                "{end_label}": "label3_end_label",
                "{inputid}": "label2_inputid",
                "open_encoding": "utf8",
                "tag2_best": "bag1",
            },
            "step0": {"tag1_test": "tag1", "comment1_det": "234"},
        },
    }

    # print(shlex.quote(json.dumps(config)))
    # parse cmd line with csv excel for rudimentary quoting, additional quoting use --json


def add_log_set_to_namespace(namespace, log_set: Tuple[str]) -> None:
    """
    Helper to add `log_set` to `argparse` `namespace`.

    Args:
        namespace: `argparse` `namespace` to add `log_set` to.
        log_set: `log_set` to add as tuple of `(name, pattern, type)`.
    """
    logsets = getattr(namespace, "logset", None)
    if logsets is None:
        logsets = []
    logsets.append(log_set)
    setattr(namespace, "logset", logsets)


def add_config_to_namespace(
    namespace, key: str, value: str, log_set_name: str = "defaults"
) -> None:
    """
    Helper to add config keys and values to `argparse` `namespace`.

    Args:
        namespace: `argparse` `namespace` to add config keys and values to.
        key: Config key to add.
        value: Config value to add.
        log_set_name: Name of `log_set` this is config for.
    """
    config = getattr(namespace, "config", None)
    if config is None:
        config = {}
    logset_config = config.setdefault(log_set_name, {})
    logset_config[key] = value
    setattr(namespace, "config", config)


def parse_config_key_values(
    config_params: List[str], namespace, log_set: Optional[str] = None
) -> None:
    """
    Parses key:value pairs for config command-line arguments.

    Args:
        config_params: List of strings as `key:value` or `key=value` to parse. 
            First `:` or `=` used to split key and value.
        namespace: `argparse` namespace to add arguments to (added as 
            `{config_name:{key1:value1...}}`.
        log_set: Name of `log_set` that this is of override config for. If `None`,
            assumed to be `defaults` config and any `config_param` without a
            `:` or `=` will set new `log_set` name for subsequent `config_params`.
            e.g. `log_set=None` and `config_params="k1:v1,k2:v2,step1,k3:v3"`
            will set `{"defaults": {"k1":"v1", "k2":"v2"}, "step1": {"k3":"v3"}}`.
            If `log_set` is not `None`, all `config_params` must be for given 
            `log_set` i.e. all `config_params` must be key:value pairs and 
            contain a `:` or `=`.

    Raises:
        ValueError: If `config_params` are of incorrect format.
    """
    if log_set is None:
        log_set = "defaults"
        new_logset_allowed = True
    else:
        new_logset_allowed = False
    for config_param in config_params:
        if not re.search("[=:]", config_param):
            if new_logset_allowed:
                log_set = config_param
                continue
            else:
                raise ValueError(
                    f"config param '{config_param}' should be 'key=value' or 'key:value'"
                )
        key, value = re.split("[=:]", config_param, 1)
        # print(f"{key} : {value}")
        add_config_to_namespace(namespace, key, value, log_set)


# format:
class ParseLogset(argparse.Action):
    """
    `argparse.Action` to parse `--logset` arguments in CSV format:
    `name,pattern,type(,config...)`, config is `key=value` or `key:value` with 
    first `=` or `:` used as split char.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        splitter = csv.reader(values, delimiter=",")
        for i, row in enumerate(splitter):
            if len(row) < 3:
                parser.error(
                    f'"{values[i]}": '
                    f"{self.dest} must be at least 3 comma-separated values"
                )
            add_log_set_to_namespace(namespace, tuple(row[:3]))
            log_set_name = row[0]
            config_params = row[3:]
            try:
                parse_config_key_values(config_params, namespace, log_set_name)
            except ValueError as e:
                parser.error(f'"{values[i]}": ' + str(e))


#
class ParseConfig(argparse.Action):
    """
    `argparse.Action` to parse `--config` arguments in CSV format:
    `log_set1,key1=value1,key2=value2,log_set2,...`
    """

    def __call__(self, parser, namespace, values, option_string=None):
        splitter = csv.reader(values, delimiter=",")
        for i, row in enumerate(splitter):
            config_params = row
            try:
                parse_config_key_values(config_params, namespace)
            except ValueError as e:
                parser.error(f'"{values[i]}": ' + str(e))


class ParseJSONConfig(argparse.Action):
    """
    `argparse.Action` to parse full config in JSON format. The following
    formats (in pseudo-JSON), or a mixture of the two, are allowed:

    ```
    {
        "log_set": [ # any key that starts with "log_set" or "logset"
            [
                name,
                pattern,
                type 
            ],
            ...
        ],
        "config": { # any key that starts with "config"
            "defaults": {
                key: value,
                ...
            },
            name: { # optional override config for log_set by name 
                key: value,
                ...
            },
        },
    
    }
    ```
    or
    ```
    {
        "log_set": [ # any key that starts with "log_set" or "logset"
            [
                name,
                pattern,
                type,
                { # allow optional 4th element as override config
                    key: value,
                    ...
                }  
            ],
            ...
        ],
        "defaults": { # top-level defaults config dict
            key: value,
            ...
        },
        
    }
    ```
    """

    def __call__(self, parser, namespace, values, option_string=None):
        if not isinstance(values, list):
            values = [values]  # values is not list if no nargs

        for value in values:

            def process_config_dict(log_set_name, config):
                # helper to process a named config dict
                if not isinstance(config, dict):
                    parser.error(
                        f'"{value}": expected {log_set_name} config "{config}" to be a dict/json object'
                    )
                for config_key, config_value in config.items():
                    add_config_to_namespace(
                        namespace, config_key, config_value, log_set_name,
                    )

            parsed = json.loads(value)
            if not isinstance(parsed, dict):
                parser.error(f'"{value}": expected top-level to be a dict/json object')
            for top_key, top_value in parsed.items():
                if top_key.startswith("log_set") or top_key.startswith("logset"):
                    if not isinstance(top_value, list):
                        parser.error(
                            f'"{value}": expected {top_key} to be a list/json array'
                        )
                    for log_set in top_value:
                        if not isinstance(log_set, list):
                            parser.error(
                                f'"{value}": expected log_set "{log_set}" to be a list/json array'
                            )
                        if len(log_set) < 3:
                            parser.error(
                                f'"{value}": expected log_set "{log_set}" to be at least 3 values'
                            )
                        if any(not isinstance(log_set[i], str) for i in range(3)):
                            parser.error(
                                f'"{value}": expected log_set "{log_set}" to be name,pattern,type strings'
                            )
                        add_log_set_to_namespace(namespace, tuple(log_set[:3]))
                        if len(log_set) > 3:  # allow optional dict as config
                            config = log_set[3]
                            log_set_name = log_set[0]
                            process_config_dict(log_set_name, config)

                            if len(log_set) > 4:
                                warning(
                                    f'"{value}": ignoring additional elements in {log_set_name} "{log_set[4:]}"'
                                )
                elif top_key.startswith("config"):
                    if not isinstance(top_value, dict):
                        parser.error(
                            f'"{value}": expected {top_key} to point to a dict/json object'
                        )
                    for config_name, config_dict in top_value.items():
                        process_config_dict(config_name, config_dict)
                elif top_key.startswith("default"):
                    # allow top level defaults dict
                    process_config_dict("defaults", top_value)
                else:
                    parser.error(f'"{value}": unexpected top-level key {top_key}')


def process_arguments(argv):
    parser = argparse.ArgumentParser(
        prog=_PROGRAM_NAME,
        description="abrije is a generic log parser and summariser.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--version",
        action="version",
        version="{} {}".format(_PROGRAM_NAME, _PROGRAM_VERSION),
    )
    parser.add_argument(
        "-v", "--verbose", action="count", help="increase verbosity (v=info, vv=debug)"
    )

    parser.add_argument(
        "-l",
        "--logset",
        action=ParseLogset,
        nargs="+",
        help="Each LOGSET as comma-separated values, Excel-style quoting allowed: "
        '"name,pattern,type(,key1=value1,...)" '
        'with "name" as log_set name, "pattern" as a snakemake-style path pattern '
        '(containing {wildcards}) to find log files to parse, and "type" '
        "specifying the parser to use for the log files. Any additional values "
        "are used as override config for this log_set (key and value pairs "
        "separated by : or =).",
    )

    parser.add_argument(
        "-c",
        "--config",
        action=ParseConfig,
        nargs="+",
        help="Config as comma-separated values, Excel-style quoting allowed. "
        'e.g. "key1=value1,key2=value2,log_set1,key3=value3" '
        'with "key1=value1" and "key2=value2" used for defaults. and '
        '"key3=value3" used for "log_set1" overrides. Key and value can be '
        "separated by : or =, first one found used.",
    )

    parser.add_argument(
        "-j",
        "--json",
        nargs="+",
        action=ParseJSONConfig,
        help="Full config as json e.g. "
        '\'{"log_set": [name1,pattern1,type1],...], '
        '"config": {"defaults": {key1:value1,...}, name1: {key2:value2,...}}}\''
        ". Note that quotes must be escaped as appropriate for command line. "
        "Can be used instead of, or in addition to, --logset and --config.",
    )

    parser.add_argument(
        "-o",
        "--output",
        action="store",
        help="output filename, will be replaced if exists; stdout if not given",
    )
    # TODO: parser import

    args = parser.parse_args(argv)
    set_verbosity(args.verbose)
    logger.info(f"Verbosity: {args.verbose}")
    logger.info(f"List of log_sets: {args.logset}")
    logger.info(f"Config: {args.config}")
    logger.info(f"Output file: {args.output}")
    return args


def execute_abrije(argv=None, return_result=False):
    # return_result = False will print output to stdout if no output file set
    # return_result = True will return output to caller instead
    if argv is None:
        argv = sys.argv[1:]  # if parameters not provided, use sys.argv
    args = process_arguments(argv)

    abrije_ = Abrije()
    abrije_.import_parsers(abrije.parsers)
    if args.logset:
        for log_set in args.logset:
            abrije_.add_log_set(*log_set)

    if args.config:
        for config_name, parameters in args.config.items():
            abrije_.add_to_config(parameters, config_name)

    abrije_.process_log_sets()
    if args.output is not None:
        abrije_.write_output(args.output)
    else:
        with io.StringIO() as out_buffer:
            abrije_.write_output(out_buffer)
            output = out_buffer.getvalue()
        if return_result:
            return output
        print(output)





if __name__ == "__main__":
    # main()
    # sys.path.append(".") # cheat for importing unpackaged abrije
    # import doctest

    # doctest.testmod(optionflags = doctest.ELLIPSIS)
    # doctest.testmod()
    execute_abrije()



#!/usr/bin/env python
# -*- coding: utf-8 -*
import json
import shlex

snakemake_rules = None

def abrij_init(rules):
    global snakemake_rules
    snakemake_rules = rules

def abrij_log_set(rule_name, optional = False):
    # get rule_name from rules, and prepare as a log_set
    # if optional = True, return None instead of throwing exception in case rule is has no abrije_ params
    global snakemake_rules
    try:
        rule_obj = getattr(snakemake_rules, rule_name)
        rule_params = getattr(rule_obj, "params")
        rule_log = getattr(rule_obj, "log")
    except AttributeError:
        raise Exception(f"abrije: rule {rule_name} does not exist or has no log/params attr")
    
    log_type = None
    rule_valid = False # set to True if any abrije_ param is found
    for param_name in dir(rule_params):
        if param_name.startswith("abrije_"):
            rule_valid = True
            if param_name=="abrije_log_type":
                log_type = getattr(rule_params, param_name)
            else:
                raise Exception(f"abrije: rule {rule_name} has unknown abrije param {param_name}") 

    if not rule_valid:
        if optional:
            return None
        raise Exception(f"abrije: rule {rule_name} has no abrije_ params")
    
    if log_type is None:
        raise Exception(f"abrije: rule {rule_name} has no abrije_log_type")
    if not rule_log or len(rule_log)>1:
        raise Exception(f"abrije: rule {rule_name} should have single log line")
    return (rule_name, rule_log[0], log_type)


def abrij_log_sets(rule_list):
    # process all rules in rule_list as log_sets
    log_sets = []
    for rule_name in rule_list:
        log_sets.append(abrij_log_set(rule_name))
    return log_sets

def _abrij_is_rule(obj):
    # returns True if obj is a snakemake rule
    # return isinstance(obj, snakemake.rules.RuleProxy)
    return hasattr(obj, "params")

def abrij_all_log_sets():
    # find all rules, and attempt to process as log_sets with optional enabled
    # any rules with no abrije_ params will be ignored
    global snakemake_rules
    log_sets = []
    for item in dir(snakemake_rules):
        obj = getattr(snakemake_rules, item)
        if _abrij_is_rule(obj):
            rule_name = item
            log_set = abrij_log_set(rule_name, optional = True)
            if log_set is not None:
                log_sets.append(log_set)
    return log_sets

def abrij_json_param(abrij_config):
    # returns json param to be used when calling abrije
    # returns as a lambda to prevent any wildcard substitution
    return lambda wildcards: shlex.quote(json.dumps(abrij_config))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pytest
import io
from collections import OrderedDict

from abrije.parsers import *
from abrije.abrije import NamedTokens


csvwide_test1 = "heading1,heading2,heading3\nvalue1,value2,value3"
csvwide_test1a = "heading1, heading2 ,heading3\n value1  ,\tvalue2\t, value3"
csvwide_test1b = " heading1 ,\theading2 ,heading3   \n value1\t,value2, value3 "
csvwide_test1c = "heading1,heading2,heading3\nvalue1,value2,value3\n"
csvwide_test1d = (
    "heading1\theading2\theading3\nvalue1\tvalue2\tvalue3"  # "parse_delimiter": "\t"
)
csvwide_test1e = "value1,value2,value3"  # "parse_no_headings" : "yes"
csvwide_expected1 = (
    ["heading1", "heading2", "heading3"],
    ["value1", "value2", "value3"],
)
csvwide_test2 = "heading1,heading2,heading3\nvalue1"
csvwide_expected2 = (
    ["heading1", "heading2", "heading3"],
    ["value1", None, None],
)
csvwide_test3 = "headingA,headingB\nvalue1,value2,value3"
csvwide_expected3 = (
    ["headingA", "headingB", "heading3"],
    ["value1", "value2", "value3"],
)
csvwide_test4 = "heading1,heading2,heading3\n"
csvwide_test4a = "heading1,heading2,heading3\n\n"
csvwide_expected4 = (
    ["heading1", "heading2", "heading3"],
    [None, None, None],
)
csvwide_test5 = "heading1,heading2,heading3\nvalue1,value2,value3\nvalue4,value5"
csvwide_expected5 = (
    [
        "heading1",
        "heading2",
        "heading3",
        "heading1_row2",
        "heading2_row2",
        "heading3_row2",
    ],
    ["value1", "value2", "value3", "value4", "value5", None],
)
csvwide_test6 = "heading1,heading2,heading3\nvalue1,value2,value3\n,\n"
csvwide_expected6 = (
    [
        "heading1",
        "heading2",
        "heading3",
        "heading1_row2",
        "heading2_row2",
        "heading3_row2",
    ],
    ["value1", "value2", "value3", "", "", None],
)
csvwide_test7 = (
    "heading1,heading2,heading3\nvalue1,value2,value3\n\n"  # extra newline skipped
)
csvwide_test7a = (
    "heading1,heading2,heading3\n\nvalue1,value2,value3\n"  # extra newline skipped
)
csvwide_expected7 = (
    ["heading1", "heading2", "heading3",],
    ["value1", "value2", "value3"],
)
csvwide_test8 = "heading1,heading2,heading3\nvalue1,value2,value3"
csvwide_expected8a = ( # "heading_prefix": "pre"
    ["preheading1", "preheading2", "preheading3"],
    ["value1", "value2", "value3"],
)
csvwide_expected8b = ( # "heading_suffix": "suf"
    ["heading1suf", "heading2suf", "heading3suf"],
    ["value1", "value2", "value3"],
)
csvwide_expected8c = ( # "value_prefix": "pre"
    ["heading1", "heading2", "heading3"],
    ["prevalue1", "prevalue2", "prevalue3"],
)
csvwide_expected8d = ( # "value_suffix": "suf"
    ["heading1", "heading2", "heading3"],
    ["value1suf", "value2suf", "value3suf"],
)
csvwide_test9 = " heading1 ,\theading2\t,heading3\n value1 ,\tvalue2\t,value3"
csvwide_expected9a = ( # default, whitespace stripped
    ["heading1", "heading2", "heading3"],
    ["value1", "value2", "value3"],
)
csvwide_expected9b = ( # "parse_heading_strip" : ""
    [" heading1 ", "\theading2\t", "heading3"],
    ["value1", "value2", "value3"],
)
csvwide_expected9c = ( # "parse_value_strip" : ""
    ["heading1", "heading2", "heading3"],
    [" value1 ", "\tvalue2\t", "value3"],
)
csvwide_expected9d = ( # "parse_heading_strip" : "h3"
    [" heading1 ", "\theading2\t", "eading"],
    ["value1", "value2", "value3"],
)
csvwide_expected9e = ( # "parse_value_strip" : "v3"
    ["heading1", "heading2", "heading3"],
    [" value1 ", "\tvalue2\t", "alue"],
)
csvwide_test10 = "heading1,heading2,heading3\nvalue1,value2"
csvwide_expected10a = ( # default: missing becomes None
    ["heading1", "heading2", "heading3"],
    ["value1", "value2", None],
)
csvwide_expected10b = ( # "parse_value_missing": "value3"
    ["heading1", "heading2", "heading3"],
    ["value1", "value2", "value3"],
)
csvwide_expected10c = ( # "value_prefix": "pre"
    ["heading1", "heading2", "heading3"],
    ["prevalue1", "prevalue2", None],
)
csvwide_expected10d = ( # parse_value_missing: "value3", "value_prefix": "pre"
    ["heading1", "heading2", "heading3"],
    ["prevalue1", "prevalue2", "prevalue3"],
)



@pytest.mark.parametrize(
    "test_input,test_tokens,expected",
    [
        (csvwide_test1, {}, csvwide_expected1),
        (csvwide_test1a, {}, csvwide_expected1),
        (csvwide_test1b, {}, csvwide_expected1),
        (csvwide_test1c, {}, csvwide_expected1),
        (csvwide_test1d, {"parse_delimiter": "\t"}, csvwide_expected1),
        (csvwide_test1e, {"parse_no_headings": "yes"}, csvwide_expected1),
        (csvwide_test2, {}, csvwide_expected2),
        (csvwide_test3, {}, csvwide_expected3),
        (csvwide_test4, {}, csvwide_expected4),
        (csvwide_test4a, {}, csvwide_expected4),
        (csvwide_test5, {}, csvwide_expected5),
        (csvwide_test6, {}, csvwide_expected6),
        (csvwide_test7, {}, csvwide_expected7),
        (csvwide_test8, {"heading_prefix": "pre"}, csvwide_expected8a),
        (csvwide_test8, {"heading_suffix": "suf"}, csvwide_expected8b),
        (csvwide_test8, {"value_prefix": "pre"}, csvwide_expected8c),
        (csvwide_test8, {"value_suffix": "suf"}, csvwide_expected8d),
        (csvwide_test9, {}, csvwide_expected9a),
        (csvwide_test9, {"parse_heading_strip" : ""}, csvwide_expected9b),
        (csvwide_test9, {"parse_value_strip" : ""}, csvwide_expected9c),
        (csvwide_test9, {"parse_heading_strip" : "h3"}, csvwide_expected9d),
        (csvwide_test9, {"parse_value_strip" : "v3"}, csvwide_expected9e),
        (csvwide_test10, {}, csvwide_expected10a),
        (csvwide_test10, {"parse_value_missing": "value3"}, csvwide_expected10b),
        (csvwide_test10, {"value_prefix": "pre"}, csvwide_expected10c),
        (csvwide_test10, {"parse_value_missing": "value3", "value_prefix": "pre"},
            csvwide_expected10d),

    ],
)
def test_parser_csv_wide(test_input, test_tokens, expected):

    expected_result = NamedTokens.fromLists(*expected)
    with io.StringIO(test_input) as handle:
        result = NamedTokens.fromLists(
            *svtextparser_csv_wide(handle, tokens=test_tokens)
        )
        assert result == expected_result


@pytest.mark.parametrize(
    "test_input,test_tokens,expected",
    [
        (csvwide_test1d, {}, csvwide_expected1),
        (csvwide_test1, {"parse_delimiter": ","}, csvwide_expected1),
        (csvwide_test1b, {"parse_delimiter": ","}, csvwide_expected1),
    ],
)
def test_parser_tsv_wide(test_input, test_tokens, expected):

    expected_result = NamedTokens.fromLists(*expected)
    with io.StringIO(test_input) as handle:
        result = NamedTokens.fromLists(
            *svtextparser_tsv_wide(handle, tokens=test_tokens)
        )
        assert result == expected_result


csvtall_test1 = "heading1,value1\nheading2,value2\nheading3,value3"
csvtall_test1a = " heading1,value1\t\nheading2\t,value2\nheading3,value3"
csvtall_test1b = "heading1  ,value1  \nheading2   ,value2\n heading3\t,value3   "
csvtall_test1c = "heading1,value1\nheading2,value2\nheading3,value3\n"
csvtall_test1d = (
    "heading1\tvalue1\nheading2\tvalue2\nheading3\tvalue3"  # "parse_delimiter": "\t"
)
csvtall_test1e = "value1\nvalue2\nvalue3"  # "parse_no_headings" : "yes"
csvtall_expected1 = (
    ["heading1", "heading2", "heading3"],
    ["value1", "value2", "value3"],
)
csvtall_test2 = "heading1,value1\nheading2\nheading3\n"
csvtall_expected2 = (
    ["heading1", "heading2", "heading3"],
    ["value1", None, None],
)
csvtall_test3 = "headingA,value1\nheadingB,value2\nvalue3"
csvtall_expected3 = (  # without "parse_indiv_as_values"
    ["headingA", "headingB", "value3"],
    ["value1", "value2", None],
)
csvtall_expected3a = (  # "parse_indiv_as_values" : "yes"
    ["headingA", "headingB", "heading3"],
    ["value1", "value2", "value3"],
)
csvtall_test4 = "heading1\nheading2\nheading3\n"
csvtall_test4a = "heading1\nheading2\nheading3\n\n"
csvtall_expected4 = (
    ["heading1", "heading2", "heading3"],
    [None, None, None],
)
csvtall_test5 = "heading1,value1,value4\nheading2,value2,value5\nheading3,value3"
csvtall_expected5 = (
    ["heading1", "heading1_row2", "heading2", "heading2_row2", "heading3"],
    ["value1", "value4", "value2", "value5", "value3"],
)
csvtall_test6 = "heading1,value1,\nheading2,value2,\nheading3,value3"
csvtall_expected6 = (
    ["heading1", "heading1_row2", "heading2", "heading2_row2", "heading3"],
    ["value1", "", "value2", "", "value3"],
)
csvtall_test7 = (
    "heading1,value1\nheading2,value2\nheading3,value3\n\n"  # extra newline skipped
)
csvtall_test7a = (
    "heading1,value1\n\nheading2,value2\nheading3,value3\n"  # extra newline skipped
)
csvtall_expected7 = (
    ["heading1", "heading2", "heading3",],
    ["value1", "value2", "value3"],
)
csvtall_test8 = "heading1,value1\nheading2,value2\nheading3,value3"
csvtall_expected8a = ( # "heading_prefix": "pre"
    ["preheading1", "preheading2", "preheading3"],
    ["value1", "value2", "value3"],
)
csvtall_expected8b = ( # "heading_suffix": "suf"
    ["heading1suf", "heading2suf", "heading3suf"],
    ["value1", "value2", "value3"],
)
csvtall_expected8c = ( # "value_prefix": "pre"
    ["heading1", "heading2", "heading3"],
    ["prevalue1", "prevalue2", "prevalue3"],
)
csvtall_expected8d = ( # "value_suffix": "suf"
    ["heading1", "heading2", "heading3"],
    ["value1suf", "value2suf", "value3suf"],
)
csvtall_test9 = " heading1 , value1 \n\theading2\t,\tvalue2\t\nheading3,value3"
csvtall_expected9a = ( # default, whitespace stripped
    ["heading1", "heading2", "heading3"],
    ["value1", "value2", "value3"],
)
csvtall_expected9b = ( # "parse_heading_strip" : ""
    [" heading1 ", "\theading2\t", "heading3"],
    ["value1", "value2", "value3"],
)
csvtall_expected9c = ( # "parse_value_strip" : ""
    ["heading1", "heading2", "heading3"],
    [" value1 ", "\tvalue2\t", "value3"],
)
csvtall_expected9d = ( # "parse_heading_strip" : "h3"
    [" heading1 ", "\theading2\t", "eading"],
    ["value1", "value2", "value3"],
)
csvtall_expected9e = ( # "parse_value_strip" : "v3"
    ["heading1", "heading2", "heading3"],
    [" value1 ", "\tvalue2\t", "alue"],
)
csvtall_test10 = "heading1,value1\nheading2,value2\nheading3"
csvtall_expected10a = ( # default: missing becomes None
    ["heading1", "heading2", "heading3"],
    ["value1", "value2", None],
)
csvtall_expected10b = ( # "parse_value_missing": "value3"
    ["heading1", "heading2", "heading3"],
    ["value1", "value2", "value3"],
)
csvtall_expected10c = ( # "value_prefix": "pre"
    ["heading1", "heading2", "heading3"],
    ["prevalue1", "prevalue2", None],
)
csvtall_expected10d = ( # parse_value_missing: "value3", "value_prefix": "pre"
    ["heading1", "heading2", "heading3"],
    ["prevalue1", "prevalue2", "prevalue3"],
)


@pytest.mark.parametrize(
    "test_input,test_tokens,expected",
    [
        (csvtall_test1, {}, csvtall_expected1),
        (csvtall_test1a, {}, csvtall_expected1),
        (csvtall_test1b, {}, csvtall_expected1),
        (csvtall_test1c, {}, csvtall_expected1),
        (csvtall_test1d, {"parse_delimiter": "\t"}, csvtall_expected1),
        (csvtall_test1e, {"parse_no_headings": "yes"}, csvtall_expected1),
        (csvtall_test2, {}, csvtall_expected2),
        (csvtall_test3, {}, csvtall_expected3),
        (csvtall_test3, {"parse_indiv_as_values": "yes"}, csvtall_expected3a),
        (csvtall_test4, {}, csvtall_expected4),
        (csvtall_test4a, {}, csvtall_expected4),
        (csvtall_test5, {}, csvtall_expected5),
        (csvtall_test6, {}, csvtall_expected6),
        (csvtall_test7, {}, csvtall_expected7),
        (csvtall_test8, {"heading_prefix": "pre"}, csvtall_expected8a),
        (csvtall_test8, {"heading_suffix": "suf"}, csvtall_expected8b),
        (csvtall_test8, {"value_prefix": "pre"}, csvtall_expected8c),
        (csvtall_test8, {"value_suffix": "suf"}, csvtall_expected8d),
        (csvtall_test9, {}, csvtall_expected9a),
        (csvtall_test9, {"parse_heading_strip" : ""}, csvtall_expected9b),
        (csvtall_test9, {"parse_value_strip" : ""}, csvtall_expected9c),
        (csvtall_test9, {"parse_heading_strip" : "h3"}, csvtall_expected9d),
        (csvtall_test9, {"parse_value_strip" : "v3"}, csvtall_expected9e),
        (csvtall_test10, {}, csvtall_expected10a),
        (csvtall_test10, {"parse_value_missing": "value3"}, csvtall_expected10b),
        (csvtall_test10, {"value_prefix": "pre"}, csvtall_expected10c),
        (csvtall_test10, {"parse_value_missing": "value3", "value_prefix": "pre"},
            csvtall_expected10d),
    ],
)
def test_parser_csv_tall(test_input, test_tokens, expected):

    expected_result = NamedTokens.fromLists(*expected)
    with io.StringIO(test_input) as handle:
        result = NamedTokens.fromLists(
            *svtextparser_csv_tall(handle, tokens=test_tokens)
        )
        assert result == expected_result


@pytest.mark.parametrize(
    "test_input,test_tokens,expected",
    [
        (csvtall_test1d, {}, csvtall_expected1),
        (csvtall_test1, {"parse_delimiter": ","}, csvtall_expected1),
        (csvtall_test1b, {"parse_delimiter": ","}, csvtall_expected1),
    ],
)
def test_parser_tsv_tall(test_input, test_tokens, expected):

    expected_result = NamedTokens.fromLists(*expected)
    with io.StringIO(test_input) as handle:
        result = NamedTokens.fromLists(
            *svtextparser_tsv_tall(handle, tokens=test_tokens)
        )
        assert result == expected_result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import OrderedDict


def parser_blank(infile) -> dict:
    """
    Parser that does nothing, for testing.
    """
    return OrderedDict()


parser_notcallable = "a string"
# 'parser' that is not a function, produces error, for testing

# expected result of importing
expected_import_parser = {"blank": ("parser_blank", "text", parser_blank)}
